"""
Models module where all of the models are specified
"""

import enum
import uuid
from typing import List
from flask_sqlalchemy import BaseQuery
from sqlalchemy import event
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import aliased
from werkzeug.security import check_password_hash, generate_password_hash

from portal import db
from portal.database.exceptions import PortalDbError
from portal.database.mixins import EntityBase
from portal.tools import time
from portal.tools.time import normalize_time


def _repr(instance) -> str:
    """Repr helper function
    Args:
        instance: Instance of the model

    Returns(str): String representation of the model
    """
    no_include = {'password_hash', 'review', 'course', 'user', 'project', 'group', 'role',
                  'review_items'}
    if 'EXCLUDED' in vars(instance.__class__):
        no_include.update(instance.__class__.EXCLUDED)
    result = f"{instance.__class__.__name__}: "
    for key, value in vars(instance).items():
        if not key.startswith("_") and key not in no_include:
            result += f"{key}={value} "
    return result


class User(db.Model, EntityBase):
    """User entity model

    Attributes:
        id: UUID string
        uco: University identification number
        email: User's email
        username: User's username (xlogin for example)
        name: User's full name
        is_admin: Whether user is administrator
        password_hash: Hashed password
        submissions: Collection of user's submissions
        review_items: Collection of review_items created by user
        roles: Collection of all roles the user has
        groups: Collection of groups the user belongs to
    """
    EXCLUDED = ['password_hash', 'submissions', 'review_items']
    __tablename__ = 'user'
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    uco = db.Column(db.Integer)
    email = db.Column(db.String(50), unique=True, nullable=False)
    username = db.Column(db.String(15), unique=True, nullable=False)
    name = db.Column(db.String(50))
    is_admin = db.Column(db.Boolean, default=False)
    # optional - after password change
    password_hash = db.Column(db.String(120), default=None)

    submissions = db.relationship("Submission", back_populates="user", cascade="all, delete-orphan",
                                  passive_deletes=True)
    review_items = db.relationship("ReviewItem", back_populates="user")

    def set_password(self, password: str):
        """Sets password for the user
        Args:
            password(str): Unhashed password
        """
        self.password_hash = generate_password_hash(password)

    def verify_password(self, password: str) -> bool:
        """Verifies user's password
        Args:
            password(str): Unhashed password

        Returns(bool): True if the password is valid, False otherwise

        """
        return check_password_hash(self.password_hash, password)

    @property
    def courses(self) -> list:
        """Gets courses the users is a part of

        Returns(list): List of courses
        """
        return self.query_courses().all()

    def query_courses(self) -> BaseQuery:
        """Queries courses the users is signed in

        Returns(BaseQuery): BaseQuery that can be executed
        """
        return Course.query.join(Course.roles).join(
            Role.users).filter(User.id == self.id)

    def get_roles_in_course(self, course: 'Course') -> list:
        """Gets user's roles for the course

        Args:
            course(Course): Course instance

        Returns(list): List of roles
        """
        return self.query_roles_in_course(course=course).all()

    def query_roles_in_course(self, course: 'Course') -> BaseQuery:
        """Queries user's roles for the course

        Args:
            course(Course): Course instance

        Returns(BaseQuery): BaseQuery that can be executed
        """
        return Role.query.filter_by(course=course) \
            .join(Role.users).filter(User.id == self.id)

    def query_groups_in_course(self, course: 'Course',
                               project: 'Project' = None) -> BaseQuery:
        """Queries user's groups for course and project

        Args:
            course(Course): Course instance
            project(Project): Project Instance (optional)

        Returns(BaseQuery): BaseQuery that can be executed
        """
        base_query = Group.query.filter_by(course=course)
        if project:
            base_query = base_query.join(Group.projects) \
                .filter((Project.id == project.id) & (course.id == Project.course_id))
        return base_query.join(Group.users).filter(User.id == self.id)

    def get_groups_in_course(self, course: 'Course',
                             project: 'Project' = None) -> list:
        """Gets user's groups for course and project (optional)

        Args:
            course(Course): Course instance
            project(Project): Project Instance (optional)

        Returns(list): List of groups
        """
        return self.query_groups_in_course(
            course=course, project=project).all()

    def query_users_in_group_based_on_role(self, course: 'Course',
                                           role: 'Role', project: 'Project' = None) -> BaseQuery:
        """Queries users for user's group's based on role

        Args:
            course(Course): Course instance
            project(Project): Project Instance
            role(Role): Role instance

        Returns(BaseQuery): BaseQuery that can be executed
        """
        groups = self.query_groups_in_course(
            course=course, project=project).subquery()
        grp_alias = aliased(Group)
        return User.query.join(grp_alias, User.groups) \
            .join(groups).join(User.roles) \
            .filter((Role.id == role.id) & (course.id == Role.course_id))

    def get_users_in_group_based_on_role(self, course: 'Course', role: 'Role',
                                         project: 'Project' = None) -> list:
        """Gets users for user's group's based on role

         Args:
             course(Course): Course instance
             project(Project): Project Instance
             role(Role): Role instance

         Returns(BaseQuery): BaseQuery that can be executed
         """
        return self.query_users_in_group_based_on_role(course=course, project=project,
                                                       role=role).all()

    def query_permissions_for_course(self, course: 'Course'):
        """Returns the query for users permissions in the course
        Args:
            course(Course): Course instance
        Returns: Query for the permissions
        """
        permissions = RolePermissions.query.join(RolePermissions.role) \
            .filter_by(course=course) \
            .join(Role.users).filter(User.id == self.id)
        return permissions

    def get_permissions_for_course(
            self, course: 'Course') -> List['RolePermissions']:
        """Gets list of permissions for the course
        Args:
            course(Course): Course instance
        Returns(list[Permissions]): List of the Permissions
        """
        return self.query_permissions_for_course(course=course).all()

    def query_projects_by_course(self, course: 'Course') -> BaseQuery:
        groups = self.get_groups_in_course(course=course)
        pids = []
        for group in groups:
            for p in group.projects:
                pids.append(p.id)

        return Project.query.filter(Project.id.in_(pids))

    def get_all_projects(self):
        result = []
        for course in self.courses:
            result += self.get_projects_by_course(course)
        return result

    def get_projects_by_course(self, course: 'Course') -> List['Project']:
        """Gets projects from the course
        Args:
            course(Course): Course instance
        Returns(List[Project]): filtered list of projects
        """
        return self.query_projects_by_course(course=course).all()

    def __init__(self, uco: int = None, email: str = None, username: str = None,
                 is_admin: bool = False) -> None:
        """Creates user instance
        Args:
            uco(int): User's UCO (school identifier)
            email(str): User's Email
            username(str): Username
            is_admin(bool): Whether user is admin
        """
        self.uco = uco
        self.email = email
        self.username = username
        self.is_admin = is_admin

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)


class Course(db.Model, EntityBase):
    """Course model
    Attributes:
        id: UUID string
        name: Human readable name of the course
        codename: Code name of the course
        notes_access_token: Access code that can be used to write to IS MUNI notes API
        roles: List of roles associated to a course
        groups: List of groups associated to a course
        projects: List of projects associated to a course
    """
    EXCLUDED = ['notes_access_token', 'roles', 'groups', 'projects']
    __tablename__ = 'course'
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    name = db.Column(db.String(50), unique=True, nullable=False)
    codename = db.Column(db.String(15), unique=True, nullable=False)
    notes_access_token = db.Column(db.String(256))
    description = db.Column(db.Text)

    roles = db.relationship("Role", back_populates="course", cascade="all, delete-orphan",
                            passive_deletes=True)
    groups = db.relationship("Group", back_populates="course", cascade="all, delete-orphan",
                             passive_deletes=True)
    projects = db.relationship("Project", back_populates="course", cascade="all, delete-orphan",
                               passive_deletes=True)

    def __init__(self, name: str = None, codename: str = None,
                 description: str = None) -> None:
        """Creates course instance
        Args:
            description (str): Course's description
            name(str): Course name
            codename(str): Course codename
        """
        self.name = name
        self.codename = codename
        self.description = description

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id

    def get_users_by_role(self, role: 'Role') -> List['User']:
        """Gets all users in the course based on their role
        Args:
            role(Role): The role to filter by
        Returns(List[User]): List of users that have the role
        """
        return User.query.join(User.roles) \
            .filter((Role.id == role.id) & (Role.course == self)).all()

    def get_users_by_group(self, group: 'Group') -> List['User']:
        """Gets all users in the group
        Args:
            group(Group): The group to find users in
        Returns(List[User]): List of users that are in the group
        """
        return User.query.join(User.groups) \
            .filter((Group.id == group.id) & (Group.course == self)).all()

    def get_users_by_group_and_role(self, group: 'Group', role: 'Role'):
        """Gets users  filtered by group and role
        Args:
            group(Group): the group to find users in
            role(Role): the role to filter by

        Returns(List[User]:

        """
        return User.query.join(User.groups) \
            .filter((Group.id == group.id) & (Group.course == self)) \
            .join(User.roles).filter((Role.id == role.id) & (Role.course == self)) \
            .all()

    def get_users_filtered(self, groups: List['Group'], roles: List['Role']):
        """Gets all users in the course who are members of at least one of
        the provided groups and have at least one of the specified roles.

        Args:
            groups(List['Group']): A list of groups. A user must be part of at least one to
                be included in the result.
            roles(List['Role']): A list of roles. A user must be part of at
                least one to be included in the result.

        Returns: Users in the course that satisfy constraints
            (are in at least one group AND have at least one role).

        """
        query = User.query.join(User.roles).filter(Role.course_id == self.id)
        if roles:
            role_ids = [role.id for role in roles]
            query = query.filter(Role.id.in_(role_ids))
        if groups:
            group_ids = [group.id for group in groups]
            query = query.join(User.groups).filter(Group.id.in_(group_ids)).filter(
                Group.course_id == self.id)
        return query.all()


class ProjectState(enum.Enum):
    """All the states in which the project can be
    """
    ACTIVE = 1
    INACTIVE = 2
    ARCHIVED = 3


class Project(db.Model, EntityBase):
    """Project model
    Attributes:
        id: UUID of the project
        name: Name of the project
        description: Description of the project
        config: Configuration of the project
        course_id: id of the course associated with the project
        course: course associated with the project
        submissions: list of submissions associated with the project
    """
    EXCLUDED = ['course', 'submissions', 'config']
    __tablename__ = 'project'
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    name = db.Column(db.String(50), nullable=False)
    description = db.Column(db.Text)
    assignment_url = db.Column(db.Text)

    config = db.relationship("ProjectConfig", back_populates="project", uselist=False,
                             cascade="all, delete-orphan", passive_deletes=True)

    course_id = db.Column(db.String(36), db.ForeignKey(
        'course.id'), nullable=False)
    course = db.relationship(
        "Course", back_populates="projects", uselist=False)
    submissions = db.relationship("Submission", back_populates="project",
                                  cascade="all, delete-orphan",
                                  passive_deletes=True)

    __table_args__ = (
        db.UniqueConstraint('course_id', 'name', name='p_course_unique_name'),
        )

    def state(self, timestamp=time.current_time()) -> ProjectState:
        """Gets project state based on the timestamp
        Args:
            timestamp: Time for which the state should be calculated
        Returns:
        """
        if not (self.config.submissions_allowed_from or
                self.config.submissions_allowed_to or
                self.config.archive_from):
            return ProjectState.INACTIVE
        if self.config.submissions_allowed_from <= timestamp < self.config.submissions_allowed_to:
            return ProjectState.ACTIVE
        elif timestamp >= self.config.archive_from:
            return ProjectState.ARCHIVED
        return ProjectState.INACTIVE

    def set_config(self, **kwargs):
        """Sets project configuration
        Keyword Args:
            test_files_source(str): Repository URL
            file_whitelist(str): Filter string for whitelist
            pre_submit_script(str): Pre submit script used in the submissions processing
            post_submit_script(str): Post submit script used in the submissions processing
            submission_parameters(str):
                Schema that defines all the parameters that should be passed to submission
            submission_scheduler_config(str):
                Configuration for the submissions scheduler
            submissions_allowed_from(time):
                Time from all the submissions are allowed from
            submissions_allowed_to(time):
                Time to which all submissions are allowed to
            archive_from(time):
                Archive all submissions from
        """
        for k, w in kwargs.items():
            if k in (
                    'test_files_source', 'file_whitelist', 'pre_submit_script',
                    'post_submit_script',
                    'submission_parameters', 'submission_scheduler_config',
                    '_submissions_allowed_from',
                    '_submissions_allowed_to', '_archive_from',
                    'submissions_allowed_from',
                    'submissions_allowed_to', 'archive_from'):
                setattr(self.config, k, w)

    def __init__(self, course: Course, name: str = None, test_files_source: str = None,
                 description: str = None):
        """Creates instance of the project
        Args:
            course(Course): Course instance
            description(des): Project's description
            name(str): Project name
            test_files_source(str): Url to test files
        """
        self.course = course
        self.name = name
        self.description = description
        self.config = ProjectConfig(
            project=self, test_files_source=test_files_source)

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class ProjectConfig(db.Model, EntityBase):
    """Project's configuration
    Attributes:
        id: UUID of the project's configuration
        project_id: Associated project's id
        project: Associated project
        test_files_source: Test files source location: git repository
        file_whitelist: whitelisted files
        pre_submit_script: Pre submit script used in the submissions processing component
        post_submit_script: Post submit script used in the submissions processing component
        submission_parameters: Post submit script used in the submissions processing component
        submission_scheduler_config: Post submit script used in the submissions scheduler
    """
    EXCLUDED = ['project']
    __tablename__ = 'projectConfig'
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    project_id = db.Column(db.String(36), db.ForeignKey(
        'project.id'), nullable=False)
    project = db.relationship(
        "Project", back_populates="config", uselist=False)

    submissions_cancellation_period = db.Column(db.Integer)
    test_files_source = db.Column(
        db.String(250), nullable=False)  # a git repository URL
    file_whitelist = db.Column(db.Text)
    pre_submit_script = db.Column(db.Text)
    post_submit_script = db.Column(db.Text)
    submission_parameters = db.Column(db.Text)
    submission_scheduler_config = db.Column(db.Text)

    _submissions_allowed_from = db.Column(db.TIMESTAMP(timezone=True))
    _submissions_allowed_to = db.Column(db.TIMESTAMP(timezone=True))
    _archive_from = db.Column(db.TIMESTAMP(timezone=True))

    @hybrid_property
    def submissions_allowed_from(self):
        """Gets from which date all the submissions are allowed from

        Returns(time): Time from submissions are allowed from

        """
        seconds = time.strip_seconds(self._submissions_allowed_from)
        return seconds

    @hybrid_property
    def submissions_allowed_to(self):
        """Gets from which date all the submissions are allowed to

        Returns(time): Time from submissions are allowed to

        """
        seconds = time.strip_seconds(self._submissions_allowed_to)
        return seconds

    @hybrid_property
    def archive_from(self):
        """Gets from which date all the submissions are archived from

        Returns(time): Time from submissions should be archived from

        """
        seconds = time.strip_seconds(self._archive_from)
        return seconds

    @submissions_allowed_from.setter
    def submissions_allowed_from(self, submissions_allowed_from):
        """Set time from which submissions are allowed to submit
        Args:
            submissions_allowed_from(time): Time from which all the submissions are allowed to be
                submitted
        """
        submissions_allowed_from = normalize_time(submissions_allowed_from)
        if self.submissions_allowed_to is not None and \
                submissions_allowed_from > self.submissions_allowed_to:
            raise PortalDbError()
        if self.archive_from is not None and \
                submissions_allowed_from > self.archive_from:
            raise PortalDbError()
        self._submissions_allowed_from = time.strip_seconds(
            submissions_allowed_from)

    @submissions_allowed_to.setter
    def submissions_allowed_to(self, submissions_allowed_to):
        """Set time from which submissions are allowed to submit
        Args:
            submissions_allowed_to(time): Time to which all the submissions are allowed to submit
        """
        submissions_allowed_to = normalize_time(submissions_allowed_to)
        if self.submissions_allowed_from is not None and \
                self.submissions_allowed_from > submissions_allowed_to:
            raise PortalDbError()
        if self.archive_from is not None and \
                self.archive_from < submissions_allowed_to:
            raise PortalDbError()
        self._submissions_allowed_to = time.strip_seconds(
            submissions_allowed_to)

    @archive_from.setter
    def archive_from(self, archive_from):
        """Set time from which submissions are allowed to archive
        Args:
            archive_from(time): Time from which all the submissions are allowed to archive
        """
        archive_from = normalize_time(archive_from)
        if self.submissions_allowed_from is not None and \
                self.submissions_allowed_from > archive_from:
            raise PortalDbError()
        if self.submissions_allowed_to is not None and \
                self.submissions_allowed_to > archive_from:
            raise PortalDbError()
        self._archive_from = time.strip_seconds(archive_from)

    def __init__(self, project: Project, test_files_source: str = None):
        """Creates instance of the project's config
        Args:
            project(Project): Project instance
            test_files_source: Source files
        """
        self.project = project
        self.test_files_source = test_files_source

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class Role(db.Model, EntityBase):
    """Role model
    Attributes:
        id: UUID of the role
        name: Name of the role
        description: Description of the role
        users: Users associated with the role
        course: Course associated with the role
        permissions: Permissions associated with the role
    """
    EXCLUDED = ['users', 'course', 'permissions']
    __tablename__ = 'role'
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    name = db.Column(db.String(50), nullable=False)
    description = db.Column(db.Text)
    users = db.relationship("User", secondary="users_roles")
    course_id = db.Column(db.String(36), db.ForeignKey(
        'course.id'), nullable=False)
    course = db.relationship("Course", back_populates="roles", uselist=False)

    permissions = db.relationship("RolePermissions", back_populates='role',
                                  cascade="all, delete-orphan",
                                  passive_deletes=True, uselist=False)

    def set_permissions(self, **kwargs):
        """Sets permissions for the role
        Args:
            **kwargs: Permissions
        """
        for k, w in kwargs.items():
            if hasattr(self.permissions, k) and k not in (
                    'id', 'role_id', 'role'):
                setattr(self.permissions, k, w)

    __table_args__ = (
        db.UniqueConstraint('course_id', 'name', name='r_course_unique_name'),
        )

    def __init__(self, course: Course, name: str = None,
                 description: str = None):
        """Creates instance of the role in a course
        Args:
            course(Course): Course instance
            description(str): Role's description
            name(str): Name of the role
        """
        self.course = course
        self.description = description
        self.name = name
        self.permissions = RolePermissions(self)

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class RolePermissions(db.Model, EntityBase):
    """Role Permissions model
    Attributes:
        id: Permissions' UUID
        role: Role associated with the permissions
    """
    EXCLUDED = ['role']
    __tablename__ = 'rolePermissions'
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    role_id = db.Column(db.String(36), db.ForeignKey(
        'role.id'), nullable=False)
    role = db.relationship("Role", back_populates="permissions", uselist=False)

    # all default to False, explicit setting via role.set_permissions is
    # required
    view_course_limited = db.Column(db.Boolean, default=False, nullable=False)
    view_course_full = db.Column(db.Boolean, default=False, nullable=False)
    update_course = db.Column(db.Boolean, default=False, nullable=False)
    handle_notes_access_token = db.Column(
        db.Boolean, default=False, nullable=False)

    assign_roles = db.Column(db.Boolean, default=False, nullable=False)
    write_roles = db.Column(db.Boolean, default=False, nullable=False)
    read_roles = db.Column(db.Boolean, default=False, nullable=False)

    write_groups = db.Column(db.Boolean, default=False, nullable=False)
    read_groups = db.Column(db.Boolean, default=False, nullable=False)

    write_projects = db.Column(db.Boolean, default=False, nullable=False)
    read_projects = db.Column(db.Boolean, default=False, nullable=False)
    archive_projects = db.Column(db.Boolean, default=False, nullable=False)

    create_submissions = db.Column(db.Boolean, default=False, nullable=False)
    resubmit_submissions = db.Column(db.Boolean, default=False, nullable=False)
    evaluate_submissions = db.Column(db.Boolean, default=False, nullable=False)

    read_submissions_all = db.Column(db.Boolean, default=False, nullable=False)
    read_submissions_groups = db.Column(
        db.Boolean, default=False, nullable=False)
    read_submissions_own = db.Column(db.Boolean, default=False, nullable=False)
    read_all_submission_files = db.Column(
        db.Boolean, default=False, nullable=False)

    read_reviews_all = db.Column(db.Boolean, default=False, nullable=False)
    read_reviews_groups = db.Column(db.Boolean, default=False, nullable=False)
    read_reviews_own = db.Column(db.Boolean, default=False, nullable=False)

    write_reviews_all = db.Column(db.Boolean, default=False, nullable=False)
    write_reviews_group = db.Column(db.Boolean, default=False, nullable=False)
    write_reviews_own = db.Column(db.Boolean, default=False, nullable=False)

    def __init__(self, role: Role):
        """Creates instance of the permissions for the role
        Args:
            role:
        """
        self.role = role

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class Group(db.Model, EntityBase):
    """Group model
    Attributes:
        id: UUID of the group
        name: Name of the Group
        course: Course associated with the Group
        users: Collection of users that are in the Group
        projects: Collection of projects that are allowed for the Group
    """
    __tablename__ = 'group'
    EXCLUDED = ['course', 'users', 'projects']
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    name = db.Column(db.String(50), nullable=False)
    description = db.Column(db.Text)
    course_id = db.Column(db.String(36), db.ForeignKey(
        'course.id'), nullable=False)
    course = db.relationship("Course", back_populates="groups", uselist=False)
    users = db.relationship("User", secondary="users_groups")
    projects = db.relationship(
        "Project", secondary="projects_groups", back_populates="groups")

    __table_args__ = (
        db.UniqueConstraint('course_id', 'name', name='g_course_unique_name'),
        )

    def __init__(self, course: Course, name: str = None,
                 description: str = None):
        """Creates instance of the group
        Args:
            course(Course): Course instance
            name(str): Name of the group
            description(str): Description of the group
        """
        self.course = course
        self.description = description
        self.name = name

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id

    def get_users_based_on_role(self, role: 'Role' = None):
        if not role:
            return self.users
        return User.query.join(User.groups).filter(Group.id == self.id) \
            .join(User.roles).filter(Role.id == role.id).all()


class SubmissionState(enum.Enum):
    """Enum of the submissions states
    """
    CREATED = 1
    READY = 2
    QUEUED = 3
    IN_PROGRESS = 4
    FINISHED = 5
    CANCELLED = 6
    ABORTED = 7
    ARCHIVED = 8


class Submission(db.Model, EntityBase):
    """Submission model
    Attributes:
        id: UUID of the submission
        scheduled_for: Date and time when the submission should be executed
        parameters: Parameters of the submission
        state: State of the submission
        note: Notes with tags for the submission

        user: User who created the submission
        project: Associated project for the submission
        review: Review associated with the submission
    """
    EXCLUDED = ['user', 'project', 'review']
    __tablename__ = 'submission'
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    scheduled_for = db.Column(db.TIMESTAMP(timezone=True))
    parameters = db.Column(db.Text, nullable=False)
    state = db.Column(
        db.Enum(SubmissionState, name='SubmissionState'), nullable=False)
    note = db.Column(db.Text)

    user_id = db.Column(db.String(36), db.ForeignKey(
        'user.id'), nullable=False)
    user = db.relationship("User", back_populates="submissions", uselist=False)
    project_id = db.Column(db.String(36), db.ForeignKey(
        'project.id'), nullable=False)
    project = db.relationship("Project", uselist=False)
    review = db.relationship("Review", back_populates="submission", cascade="all, delete-orphan",
                             passive_deletes=True, uselist=False)

    storage_task_id = db.Column(db.String(length=36))
    process_task_id = db.Column(db.String(length=36))

    def change_state(self, new_state):
        # open to extension (state transition validation, ...)
        self.state = new_state

    def __init__(self, user: User, project: Project, parameters: str = None,
                 review: 'Review' = None):
        """Creates a new submission instance
        Args:
            user(User): User instance
            project(Project): Project instance
            parameters(str): Parameters for the submission
            review(Review): Review instance
        """
        self.user = user
        self.project = project
        self.review = review
        self.parameters = parameters
        self.state = SubmissionState.CREATED

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class Review(db.Model, EntityBase):
    """Review model
    Attributes:
        id: UUID of the review
        submission: Associated submission for the review
        review_items: Review items (Comments) associated to review and submission
    """
    EXCLUDED = ['user', 'review_items', 'submission']
    __tablename__ = 'review'
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    submission_id = db.Column(db.String(36), db.ForeignKey(
        'submission.id'), nullable=False)
    submission = db.relationship(
        "Submission", uselist=False, back_populates="review")
    review_items = db.relationship("ReviewItem", back_populates="review",
                                   cascade="all, delete-orphan", passive_deletes=True)

    def __init__(self, submission: Submission):
        """Creates a review for the submission
        Args:
            submission(Submission):
        """
        self.submission = submission

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class ReviewItem(db.Model, EntityBase):
    """Review item model
    Attributes:
        id: UUID of the submission
        review: Associated review
        user: Author of the review
        content: Content of the review (normal text)
        file: File path
        line: line in the file
    """
    __tablename__ = 'reviewItem'
    EXCLUDED = ['user', 'review']
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    review_id = db.Column(db.String(36), db.ForeignKey(
        'review.id'), nullable=False)
    review = db.relationship("Review", uselist=False,
                             back_populates="review_items")
    user_id = db.Column(db.String(36), db.ForeignKey('user.id'))
    user = db.relationship("User", uselist=False,
                           back_populates="review_items")  # the review item's author
    content = db.Column(db.Text)
    file = db.Column(db.String(256), nullable=False)
    line = db.Column(db.Integer, nullable=False)

    def __init__(self, user: User, review: Review,
                 file: str, line: int, content: str):
        """Creates a review item in the review
        Args:
            user(User): The author of the review item
            review(Review): Review the item belongs to
            file(str): File the review item binds to
            line(int): Line of the file
            content(str): Review item content (text)
        """
        self.review = review
        self.user = user
        self.file = file
        self.line = line
        self.content = content

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


class Component(db.Model, EntityBase):
    """Component model:
    Attributes:
        id: UUID of the submission
        name: Name of the component
        secret: Component's associated secret
        ip_address: Ip address of the component
        type: type of the component
        nodes: Associated notes to a component
    """
    __tablename__ = 'component'
    EXCLUDED = ['secret']
    id = db.Column(db.String(length=36), default=lambda: str(
        uuid.uuid4()), primary_key=True)
    name = db.Column(db.String(30), nullable=False, unique=True)
    secret = db.Column(db.String(250))
    ip_address = db.Column(db.String(50))
    type = db.Column(db.String(25))
    notes = db.Column(db.Text)

    def __init__(self, name: str = None, secret: str = None, ip_address: str = None,
                 component_type: str = None, notes: str = None):
        """Creates component
        Args:
            name(str): Name of the component
            secret(str): Component's secret key
            ip_address(str): Component's IP address
            component_type(str): Type of the component
            notes(str): Additional notes for the component
        """
        self.name = name
        self.secret = secret
        self.ip_address = ip_address
        self.notes = notes
        self.type = component_type

    def __repr__(self):
        return _repr(self)

    def __eq__(self, other):
        return self.id == other.id


# source: http://docs.sqlalchemy.org/en/latest/orm/events.html
@event.listens_for(Submission.user_id, 'set', named=True)
def receive_set(**kw):
    submission = kw['target']
    value = kw['value']

    if submission.user_id is not None and submission.user_id != value:
        raise PortalDbError()


# Associations - must come after definitions of the affected classes
users_groups = db.Table("users_groups", db.Model.metadata,
                        db.Column("user_id", db.String(36),
                                  db.ForeignKey('user.id')),
                        db.Column("group_id", db.String(36), db.ForeignKey('group.id')))

users_roles = db.Table("users_roles", db.Model.metadata,
                       db.Column("user_id", db.String(36),
                                 db.ForeignKey('user.id')),
                       db.Column("role_id", db.String(36), db.ForeignKey('role.id')))

projects_groups = db.Table("projects_groups", db.Model.metadata,
                           db.Column("project_id", db.String(36),
                                     db.ForeignKey('project.id')),
                           db.Column("group_id", db.String(36), db.ForeignKey('group.id')))

User.roles = db.relationship("Role", secondary="users_roles")
User.groups = db.relationship("Group", secondary="users_groups")
Project.groups = db.relationship(
    "Group", secondary="projects_groups", back_populates='projects')
