import json
from datetime import timedelta
from flask_jwt_extended import create_access_token

from portal.database.models import Course, Project, User
from portal.tools.time import current_time, strip_seconds
from . import utils


def test_list(client):
    cpp = Course.query.filter_by(codename="TestCourse2").first()
    cpp_projects = len(cpp.projects)

    response = utils.make_request(
        client, f'/courses/{cpp.codename}/projects', 'get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    projects = utils.extract_data(response)
    assert len(projects) == cpp_projects
    cpp_updated = Course.query.filter_by(codename="TestCourse1").first()
    for p in projects:
        utils.assert_project_in(cpp_updated.projects, p)


def test_create(client):
    cpp = Course.query.filter_by(codename="TestCourse1").first()
    cpp_projects = len(cpp.projects)

    request_dict = {
        "name": "new_project",
        "config": {
            "test_files_source": "foo@git.git",
        },
    }
    request_json = json.dumps(request_dict)
    response = utils.make_request(client, f"/courses/{cpp.codename}/projects", method='post',
                                  data=request_json,
                                  headers={"content-type": "application/json"})
    assert response.status_code == 201
    assert response.mimetype == 'application/json'

    new_project = utils.extract_data(response)
    cpp_updated = Course.query.filter_by(codename="TestCourse1").first()
    assert len(cpp_updated.projects) == cpp_projects + 1
    assert new_project['name'] == "new_project"
    assert new_project['id']
    assert new_project['config']['test_files_source'] == "foo@git.git"
    utils.assert_project_in(cpp_updated.projects, new_project)


def test_read(client):
    cpp = Course.query.filter_by(codename="TestCourse1").first()
    p = cpp.projects[0]
    response = utils.make_request(client, f"/courses/{cpp.codename}/projects/{p.name}",
                                  method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    project = utils.extract_data(response)

    utils.assert_project(p, project)


def test_update(client):
    cpp = Course.query.filter_by(codename="TestCourse1").first()
    p = cpp.projects[0]
    p_name = p.name
    request_dict = dict(
        name="new project name",
    )
    request_json = json.dumps(request_dict)
    response = utils.make_request(client, f"/courses/{cpp.codename}/projects/{p.name}",
                                  data=request_json, method='put',
                                  headers={"content-type": "application/json"})
    assert response.status_code == 204
    assert response.mimetype == 'application/json'
    assert Project.query.filter(Project.course_id == cpp.id).filter(
        Project.name == "new project name").first()
    assert not Project.query.filter(Project.course_id == cpp.id).filter(
        Project.name == p_name).first()


def test_delete(client):
    cpp = Course.query.filter_by(codename="TestCourse1").first()
    p = cpp.projects[0]
    p_name = p.name
    cpp_projects = len(cpp.projects)
    response = utils.make_request(client, f"/courses/{cpp.codename}/projects/{p.name}",
                                  method="delete")

    assert response.status_code == 204
    assert response.mimetype == 'application/json'
    cpp_updated = Course.query.filter_by(codename="TestCourse1").first()
    assert len(cpp_updated.projects) == cpp_projects - 1
    assert not Project.query.filter(Project.course_id == cpp.id).filter(
        Project.name == p_name).first()


def test_config_read(client):
    cpp = Course.query.filter_by(codename="TestCourse1").first()
    p = cpp.projects[0]

    response = utils.make_request(client, f"/courses/{cpp.codename}/projects/{p.name}/config",
                                  method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    data = utils.extract_data(response)
    utils.assert_project_config(p.config, data)


def test_config_update(client):
    cpp = Course.query.filter_by(codename="TestCourse1").first()
    p = cpp.projects[0]
    p_name = p.name

    new_time = current_time() + timedelta(days=2)
    request_dict = dict(
        test_files_source="new source",
        pre_submit_script="a python script",
        submissions_allowed_to=new_time
    )
    request_json = json.dumps(request_dict, cls=utils.DateTimeEncoder)

    response = utils.make_request(client, f"/courses/{cpp.codename}/projects/{p.name}/config",
                                  data=request_json,
                                  headers={"content-type": "application/json"}, method='put')
    assert response.status_code == 204
    assert response.mimetype == 'application/json'

    p_updated = Project.query.filter(
        Project.course_id == cpp.id).filter_by(name=p_name).first()
    assert p_updated.config.test_files_source == "new source"
    assert p_updated.config.pre_submit_script == "a python script"
    assert p_updated.config.submissions_allowed_to == strip_seconds(new_time)


def test_list_submissions(client):
    cpp = Course.query.filter_by(codename="TestCourse1").first()
    p = cpp.projects[0]

    user = User.query.filter_by(username="student2").first()
    access_token = create_access_token(identity=user.username)
    response = utils.make_request(client, f"/courses/{cpp.codename}/projects/{p.name}/submissions",
                                  headers={"Authorization": f"Bearer {access_token}"}, method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'
    submissions = utils.extract_data(response)
    assert len(submissions) == len(p.submissions)
    for s in submissions:
        utils.assert_submission_in(p.submissions, s)


def test_list_submissions_filter(client):
    cpp = Course.query.filter_by(codename="TestCourse1").first()
    p = cpp.projects[0]
    user = User.query.filter_by(username="student2").first()
    user_submissions = len(
        [submission for submission in user.submissions if submission.project.course == cpp])
    access_token = create_access_token(identity=user.username)
    response = utils.make_request(
        client,
        f"/courses/{cpp.codename}/projects/{p.name}/submissions?user=student2",
        headers={"Authorization": f"Bearer {access_token}"}, method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'
    submissions = utils.extract_data(response)
    assert len(submissions) == user_submissions
    for s in submissions:
        utils.assert_submission_in(p.submissions, s)
        assert s['user']['id'] == user.id


def test_create_submission(client):
    cpp = Course.query.filter_by(codename="TestCourse1").first()
    p = cpp.projects[0]
    p_submissions = len(p.submissions)

    request_dict = {
        "project_params": "data for Kontr",
        "file_params": {
            "source": {
                "type": "git",
                "url": "https://gitlab.fi.muni.cz/xkompis/test-hello-world.git",
                "branch": "master",
                "checkout": "master"
            }
        }
    }
    request_json = json.dumps(request_dict, cls=utils.DateTimeEncoder)

    path = f"/courses/{cpp.codename}/projects/{p.name}/submissions"
    response = utils.make_request(
        client, path, data=request_json, method='post')

    assert response.status_code == 201
    assert response.mimetype == 'application/json'

    new_submission = utils.extract_data(response)
    p_updated = Project.query.filter(
        Project.course_id == cpp.id).filter_by(name=p.name).first()
    user_updated = User.query.filter_by(username="admin").first()

    assert len(p_updated.submissions) == p_submissions + 1
    utils.assert_submission_in(p_updated.submissions, new_submission)
    utils.assert_submission_in(user_updated.submissions, new_submission)
